BYTE_UNIT_STEP = 10

BYTES_PER_KIBIBYTE = 1 << (1 * BYTE_UNIT_STEP)
BYTES_PER_MEBIBYTE = 1 << (2 * BYTE_UNIT_STEP)
BYTES_PER_GIBIBYTE = 1 << (3 * BYTE_UNIT_STEP)

MEBIBYTES_PER_GIBIBYTE = 1 << (1 * BYTE_UNIT_STEP)

BYTES_PER_KILOBYTE = 1e3
BYTES_PER_MEGABYTE = 1e6
BYTES_PER_GIGABYTE = 1e9
BYTES_PER_TERABYTE = 1e12


SECONDS_PER_MINUTE = 60
SECONDS_PER_HOUR = 60 * SECONDS_PER_MINUTE
SECONDS_PER_DAY = 24 * SECONDS_PER_HOUR
